import os
import shutil
import typer

from collections import namedtuple
from repotracer.commands import run
from repotracer.lib import git, config
from repotracer.lib.measurement import all_measurements, ParamOption
from rich import print
from rich.console import Console
from typing import Optional, List
from typing_extensions import Annotated
import click
import questionary
from enum import Enum

from urllib.parse import urlparse


def add_repo(url_or_path: str, name: Optional[str] = None):
    console = Console()
    repo_config = git.download_repo(url_or_path)
    config.add_repo(repo_config=repo_config)
    return repo_config
    # optionally ask the user if they want to add any stats for this repo
    # and call add_stat() if they do


qstyle = questionary.Style(
    [
        (
            "highlighted",
            "reverse bold",
        )  # pointed-at choice in select and checkbox prompts
    ]
)


def add_stat(
    repo_name: Annotated[Optional[str], typer.Argument()] = None,
    stat_name: Annotated[Optional[str], typer.Argument()] = None,
):
    if repo_name is None:
        repo_name = questionary.select(
            "Which repo do you want to add a new stat for?",
            choices=config.list_repos() + ["<new repo>"],
            style=qstyle,
            qmark="🕵️",
        ).ask()
        if repo_name == "<new repo>":
            url_or_path = questionary.text(
                "Enter either a repo URL or paste a path to a local repo on your machine (we will make copy it into the repos folder):"
            ).ask()
            added_config = add_repo(url_or_path)
            repo_name = added_config.name

    if stat_name is None:
        stat_name = questionary.text("What name do you want to give the stat?").ask()
    if stat_name in config.list_stats_for_repo(repo_name):
        print(
            f"The stat '{stat_name}' already exists for the repo '{repo_name}'. Please choose a different name."
        )
        return

    # TODO make this be able to be given from the command line, for now this command
    # will require user input

    stat_type = questionary.select(
        "What type of stat do you want to add?",
        choices=all_measurements.keys(),
        style=qstyle,
        # qmark="?"
    ).ask()

    stat_description = questionary.text(
        "Give your stat a description, to explain what it does (will be used as the title of generated graphs):",
    ).ask()

    stat_params = promt_build_stat(stat_type)
    print("Building the following stat:")
    print(stat_params)
    stat_config = config.StatConfig(
        name=stat_name,
        description=stat_description,
        type=stat_type,
        path_in_repo=stat_params.pop("path_in_repo", None),
        start=stat_params.pop("start", None),
        params=stat_params.pop("params", None),
    )
    config.add_stat(repo_name, stat_config)
    run_now = questionary.confirm("Do you want to run this stat now?").ask()
    if run_now:
        run.run(repo_name, stat_name)


def promt_build_stat(stat_type: str):
    common_stat_options = [
        # Todo the required ones like type and description should just be done in the other function, or moved into here
        # ParamOption(
        #     name="description",
        #     description="A description of the stat, used in the title of generated graphs",
        #     required=True,
        # ),
        ParamOption(
            name="start",
            description="The start date for the stat, if you don't want to start at the beginning of the repo",
            required=False,
        ),
        ParamOption(
            name="path_in_repo",
            description="The path in the repo to run the stat on",
            required=False,
        ),
    ]
    # Todo move this into the definition of the measurement/stat type
    stat_options = all_measurements[stat_type].params
    required_stat_params = [option for option in stat_options if option.required]
    optional_stat_params = [option for option in stat_options if not option.required]
    stat_params = prompt_required_options(required_stat_params)
    stat_params |= prompt_for_options(
        f"The type {stat_type} has the following options. Would you like to set any of these?",
        optional_stat_params,
    )
    common_options = prompt_for_options(
        f"Additionally, stats can have the following options. Would you like to set any of these?",
        common_stat_options,
    )
    return {**common_options, "params": stat_params}


def prompt_required_options(options: List[ParamOption]):
    choices = {}
    for option in options:
        choices[option.name] = prompt_for_single_param_option(option)
    return choices


def find(list, predicate):
    for item in list:
        if predicate(item):
            return item
    return None


def prompt_for_options(prompt: str, options: List[ParamOption]):
    options_chosen = {}
    options_remaining = {option.name for option in options}
    while options_remaining:
        option_name = questionary.select(
            prompt,
            choices=list(options_remaining) + ["<none>"],
            style=qstyle,
            qmark="📈️",
        ).ask()
        if option_name == "<none>":
            break
        option_value = prompt_for_single_param_option(
            find(options, lambda x: x.name == option_name)
        )
        options_chosen[option_name] = option_value
        options_remaining.remove(option_name)
    return options_chosen


def prompt_for_single_param_option(option: ParamOption):
    return questionary.text(
        f"Please choose a value for '{option.name}': {option.description}."
    ).ask()



from dataclasses import dataclass, asdict
from typing import Any, Optional
import dacite
import os
import json5
from pathlib import Path


@dataclass()
class StatConfig(object):
    name: str
    description: str
    type: str
    params: Any
    path_in_repo: Optional[str]
    start: Optional[str]


@dataclass()
class RepoConfig(object):
    name: Optional[str]
    source: str
    # used to override the location where the repo is stored
    # for the eventuality that two repos have the same name
    storage_path: Optional[str] = None
    # defaults to master if not set
    default_branch: str | None = None
    stats: dict[str, StatConfig] | None = None


@dataclass()
class GlobalConfig(object):
    repo_storage_location: Optional[str]
    stat_storage: Optional[dict[str, Any]]
    repos: dict[str, RepoConfig]


# A dict version of default_config:
default_config_dict = {
    "stat_storage": {
        "type": "csv",
    },
    "repos": {},
}
# a dataclass version of default_config:
default_config = dacite.from_dict(GlobalConfig, default_config_dict)


# We define a ROOT_DIR for repotracer to store its data in.
# This root dir will either be $PWD/.repotracer if it exists, or $HOME/.repotracer if it doesn't.
# By default, repotracer will store 3 kinds of data:
# 1. Repos, under $HOME/.repotracer/repos. You can change this with the repo_storage_location config key.
# 2. The config file, under $ROOT_DIR/config.json.
# 3. The stats, under $ROOT_DIR/.repotracer/stats/<repo_name>/<stat_name>.{csv,png}
# Possibly later these will be configurable differently.


GLOBAL_ROOT_DIR = os.environ.get("REPOTRACER_ROOT_DIR") or os.environ.get(
    os.path.join("XDG_CONFIG_HOME", "repotracer"),
    os.path.expanduser("~/.repotracer"),
)


def get_root_dir():
    if os.path.exists(os.path.join(os.getcwd(), ".repotracer")):
        return os.path.join(os.getcwd(), ".repotracer")
    return GLOBAL_ROOT_DIR


ROOT_DIR = get_root_dir()


def get_config_path():
    if os.path.exists(os.path.join(os.getcwd(), "config.json")):
        return os.path.join(os.getcwd(), "config.json")
    return os.path.join(ROOT_DIR, "config.json")


def get_repos_dir():
    return read_config_file().repo_storage_location or os.path.join(
        GLOBAL_ROOT_DIR, "repos"
    )


def get_stats_dir():
    return read_config_file().stat_storage.get("path") or os.path.join(
        ROOT_DIR, "stats"
    )


def get_repo_storage_path(repo_name):
    repo_config = get_repo_config(repo_name)
    match repo_config.storage_path:
        case None:
            return os.path.join(get_repos_dir(), repo_config.name)
        case path:
            return Path(path).expanduser().resolve()


config_file_contents = None


def read_config_file() -> GlobalConfig:
    global config_file_contents
    if config_file_contents is None:
        try:
            print("Looking for config file at", os.path.abspath(get_config_path()))
            with open(get_config_path()) as f:
                config_file_contents = json5.load(
                    f
                )  # python 3.9 operator for dict update
        except FileNotFoundError:
            print(
                f"Could not find config file at {get_config_path()}, writing a default config file there."
            )
            write_config_file(default_config)
            config_file_contents = default_config_dict
    combined_config = default_config_dict | config_file_contents
    apply_implicit_names(combined_config)  # mutates in place
    return dacite.from_dict(GlobalConfig, combined_config)


def apply_implicit_names(config_dict: dict[str, Any]):
    for repo in config_dict.get("repos", {}).values():
        for name, stat_obj in (repo.get("stats") or {}).items():
            if "name" not in stat_obj:
                stat_obj["name"] = name


def list_repos() -> list[str]:
    try:
        return list(read_config_file().repos.keys())
    except KeyError:
        return []


def list_stats_for_repo(repo_name) -> list[str]:
    try:
        return list((read_config_file().repos[repo_name].stats or {}).keys())
    except KeyError:
        return []


def get_repo_config(repo_name: str) -> RepoConfig:
    config_data = read_config_file()
    try:
        repo_config: RepoConfig = config_data.repos[repo_name]
    except KeyError:
        known_repos = ",".join(config_data.repos.keys())
        raise Exception(
            f"Repo '{repo_name}' was not found in the config. It contains these repos: '{known_repos}'"
        )
    return repo_config


def get_stat_config(repo_name, stat_name) -> (RepoConfig, StatConfig):
    repo_config = get_repo_config(repo_name)
    try:
        stat_config = repo_config.stats[stat_name]
    except KeyError:
        valid_stats = ", ".join(repo_config.stats.keys())
        raise Exception(
            f"The stat '{stat_name}' does not exist in the config for the repo '{repo_name}'. The known stats are: '{valid_stats}'"
        )

    return repo_config, stat_config


def write_config_file(config):
    global config_file_contents
    config_path = get_config_path()
    config_dir = os.path.dirname(config_path)
    if not os.path.exists(config_dir):
        os.makedirs(config_dir, exist_ok=True)
    # Don't write None's into the config file
    to_write = asdict(
        config, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}
    )
    with open(config_path, "w") as f:
        json5.dump(to_write, f, indent=4, quote_keys=True)
    config_file_contents = to_write


def add_repo(repo_config: RepoConfig):
    config = read_config_file()
    config.repos[repo_config.name] = repo_config
    write_config_file(config)


def add_stat(repo_name: str, stat_config: StatConfig):
    config = read_config_file()
    repo_config = config.repos[repo_name]
    if not repo_config.stats:
        repo_config.stats = {}
    repo_config.stats[stat_config.name] = stat_config
    write_config_file(config)



import typer
from repotracer.commands import run, add, install_repos
from repotracer.lib.config import read_config_file


app = typer.Typer(no_args_is_help=True)

app.command()(add.add_repo)
app.command()(add.add_stat)
app.command()(install_repos.install_repos)
app.command()(run.run)


@app.command()
def print_config():
    import json
    from repotracer.lib.config import read_config_file

    print(json.dumps(read_config_file(), indent=4))


if __name__ == "__main__":
    read_config_file()
    app()



from . import git

import subprocess
import json


def script(cmd):
    return lambda: subprocess.check_output(cmd, shell=True)


def script_now(cmd):
    try:
        return subprocess.check_output(cmd, shell=True).decode("utf-8")
    except subprocess.CalledProcessError as e:
        # ignoGenericre exit code 1
        if e.returncode == 1:
            return e.output
        print("Error in script_now")
        print(e)
        print(e.output)
        raise e


def float_or_int(s):
    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except Exception as e:
            return float("NaN")


def script_auto(cmd, return_type):
    output = script_now(cmd)
    if return_type == "number":
        return {"output": float_or_int(output)}
    elif return_type == "json":
        return json.loads(output)
    else:
        raise ValueError(f"Unknown return type {return_type}")


def ripgrep_count_file(pattern):
    return int(script(f"rg -l {pattern} | wc -l"))


def rg_count(pattern: str, rg_args: str) -> int:
    filenames_with_counts = script_now(f"rg '{pattern}' --count {rg_args or ''}")
    res = {
        "total": sum(
            int(line.split(":")[-1]) for line in filenames_with_counts.splitlines()
        )
    }
    return res


def fd_count(pattern: str, extra_cli_args: str) -> int:
    filenames_with_counts = script_now(
        f"fd --glob '{pattern}' --type file {extra_cli_args or ''}"
    )
    # todo: eventually, store the individual file names as intermediate results
    res = {"total": len(filenames_with_counts.splitlines())}
    return res


def tokei_count(languages: str, extra_cli_args: str) -> int:
    if languages == "all":
        languages = ""
    language_arg = f"--t {languages}" if languages else ""
    json_out = script_now(f"tokei --output json {language_arg} {extra_cli_args or ''}")
    parsed = json.loads(json_out)
    code_totals = {language: data["code"] for language, data in parsed.items()}
    breakdown_languages = True
    if breakdown_languages:
        del code_totals["Total"]
        return code_totals
    else:
        return code_totals["Total"]


def jsx_to_tsx():
    return script_now("tokei --output json --languages jsx,tsx,js,ts")



import os
import shutil
import typer

from collections import namedtuple
from repotracer.commands import run
from repotracer.lib import git, config
from repotracer.lib.measurement import all_measurements, ParamOption
from rich import print
from rich.console import Console
from typing import Optional, List
from typing_extensions import Annotated
import click
import questionary
from enum import Enum

from urllib.parse import urlparse


def install_repos(repos: List[str] = typer.Argument(None)):
    # get the list of repos from the config
    # check to see the installed repos
    # if the repo is not installed, install it
    # todo ideally nobody outside of config should directly call read_config_file

    repos_to_install = repos or config.list_repos()
    for repo_name in repos_to_install:
        repo = config.get_repo_config(repo_name)
        if not os.path.exists(config.get_repo_storage_path(repo_name)):
            git.download_repo(repo.source)
    return



import matplotlib
import matplotlib.pyplot as plt
from matplotlib.dates import MonthLocator, ConciseDateFormatter, AutoDateFormatter
import seaborn as sns
import os
from .config import get_stats_dir
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())


def rotate(l, n):
    return l[n:] + l[:n]


sns.set_theme()
sns.set_style("whitegrid")
sns.set_palette(rotate(sns.color_palette("deep"), 2))
plt.rcParams["figure.dpi"] = 140
plt.rcParams["figure.figsize"] = (12.8, 9.6)


def plot(repo_name, stat_name, stat_description, df, run_at):
    image_path = os.path.join(get_stats_dir(), f"{repo_name}/{stat_name}.png")
    ax = df.plot()
    last_date = df.index.values[-1]

    if len(df.columns) == 1:
        column_name = df.columns[0]
        last_value = df.iloc[-1][column_name]
        ax.annotate(last_value, (last_date, last_value))

    plt.xlabel("Date")
    plt.suptitle(stat_description, y=0.93, size="xx-large", weight="semibold")
    plt.title(
        f"{repo_name}:{stat_name} by repotracer at {run_at.strftime('%Y-%m-%d %H:%M:%S')}",
        loc="right",
        fontsize="small",
        weight="light",
        alpha=0.5,
    )
    # ax.xaxis.set_major_locator(MonthLocator())
    # todo set a formatter that displays either single letter month or full month name
    # ax.xaxis.set_major_formatter(ConciseDateFormatter(MonthLocator()))

    plt.savefig(image_path, bbox_inches="tight")
    logger.info(f"Plotting {stat_name} to {image_path}")



import pandas as pd
from . import git
from tqdm.auto import tqdm
from datetime import datetime, date

from .config import RepoConfig, StatConfig, get_repos_dir, get_repo_storage_path
from .measurement import Measurement, all_measurements
from .storage import Storage, CsvStorage
from .plotter import plot
from typing import Callable
from dataclasses import dataclass
import os
import logging
from rich import print

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())
# logger.addHandler(logging.StreamHandler())o
# logging.config.dictConfig(
#     {
#         "disable_existing_loggers": True,
#     }
# )


@dataclass
class AggConfig(object):
    time_window: str = "D"
    agg_fn: Callable | None = None
    agg_window: str | None = None


def agg_percent(self):
    return self.sum() / len(self)


@dataclass()
class Stat(object):
    repo_config: RepoConfig
    stat_name: str
    start: str | None = None
    end: str | None = None
    measurement: Measurement | None = None
    agg_config: AggConfig | None = None
    path_in_repo: str | None = None

    def __init__(self, repo_config: RepoConfig, stat_params: StatConfig):
        self.repo_config = repo_config
        self.measurement = all_measurements[stat_params.type].obj(stat_params.params)
        self.stat_name = stat_params.name
        self.description = stat_params.description
        self.path_in_repo = stat_params.path_in_repo
        self.start = stat_params.start

    def cd_to_repo_and_setup(self, repo_path):
        logger.debug(f"cd from {os.getcwd()} to {repo_path}")
        os.chdir(repo_path)
        # todo this is slow on large repos
        # maybe only do it if there are untracked files, or do it manually
        # git.clean_untracked()
        git.reset_hard("HEAD")
        branch = self.repo_config.default_branch or "master"
        git.checkout(branch)
        git.pull(branch)
        if self.path_in_repo:
            os.chdir(self.path_in_repo)

    def loop_through_commits_and_measure(self, commits_to_measure):
        commit_stats = []
        # We assume we have already cd'd to the right place to measure the stat
        stat_measuring_path = os.getcwd()
        for commit in (
            pbar := tqdm(
                commits_to_measure.itertuples(index=True), total=len(commits_to_measure)
            )
        ):
            pbar.set_postfix_str(commit.Index.strftime("%Y-%m-%d"))
            if not commit.sha:
                commit_stats.append({"date": commit.Index})
                continue
            git.reset_hard(commit.sha)
            os.chdir(stat_measuring_path)
            stat = {
                "sha": commit.sha,
                "date": commit.Index,
                **self.measurement(),
            }
            commit_stats.append(stat)

        return (
            pd.DataFrame(commit_stats)
            .ffill()
            .set_index("date")
            .tz_localize(None)
            .convert_dtypes()
        )

    def run(self):
        previous_cwd = os.getcwd()
        repo_name = self.repo_config.name
        repo_path = get_repo_storage_path(repo_name)
        if not git.is_repo_setup(repo_path):
            # todo maybe don't try to download it, just error or tell them to run repotracer add repo
            raise Exception(
                f"Repo '{repo_name}' not found at {repo_path}. Run `repotracer add repo {repo_name}` to download it."
            )

        existing_df = CsvStorage().load(self.repo_config.name, self.stat_name)
        end = datetime.today()
        agg_config = self.agg_config or AggConfig(
            time_window="D", agg_fn=None, agg_window=None
        )

        self.cd_to_repo_and_setup(repo_path)
        start = self.find_start_day(existing_df)
        commits_to_measure = git.build_commit_df(start, end, agg_config.time_window)
        if len(commits_to_measure) == 0:
            logger.info(f"No commits found in the time window {start}-{end},  skipping")
            os.chdir(previous_cwd)
            return
        logger.info(f"Going from {start} to {end}, {len(commits_to_measure)} commits")
        new_df = self.loop_through_commits_and_measure(commits_to_measure)

        if agg_config.agg_fn:
            new_df.groupby(
                pd.Grouper(key="created_at", agg_freq=agg_freq), as_index=False
            ).agg(agg_config.agg_fn)

        if existing_df is not None:
            df = new_df.combine_first(existing_df)
        else:
            df = new_df

        os.chdir(previous_cwd)
        CsvStorage().save(self.repo_config.name, self.stat_name, df)
        plot(
            self.repo_config.name,
            self.stat_name,
            self.description,
            df,
            run_at=datetime.now(),
        )

    def find_start_day(self, df) -> date:
        # We need to ask the storage engine for the current version of the data
        # It should give us a df, and we can use that to find the latest days missing
        if df is None or df.empty:
            if self.start:
                start = pd.to_datetime(self.start)
                logger.debug(f"Using given self.start: {self.start}")
            else:
                first_commit = git.first_commit_date()
                logger.debug(f"Found first commit date {first_commit}")
                start = first_commit
            logger.info(
                f"No existing data found, starting from the beginning on {start}"
            )
        else:
            start = df.index.max() - pd.Timedelta(days=1)
            logger.debug(f"Found existing data date {start}")
        # Return a list of days missing
        return start



from repotracer.lib.stat import Stat
from repotracer.lib.config import get_stat_config, list_repos, list_stats_for_repo
from typing import Optional
from typing_extensions import Annotated
import typer
from rich import print
from rich.console import Console
from rich.table import Table
from dataclasses import dataclass

import os


def run(
    repo_name: Annotated[Optional[str], typer.Argument()] = None,
    stat_name: Annotated[Optional[str], typer.Argument()] = None,
    since: Optional[str] = None,
):
    pairs_to_run = []
    match (repo_name, stat_name):
        case (None, None):
            pairs_to_run = [
                (repo, stat)
                for repo in list_repos()
                for stat in list_stats_for_repo(repo)
            ]
        case (repo_name, None):
            pairs_to_run = [
                (repo_name, stat) for stat in list_stats_for_repo(repo_name)
            ]
        case (repo_name, stat_name):
            pairs_to_run = [(repo_name, stat_name)]
        case _:
            logging.error("This combination of repo_name and stat_name is invalid.")
            return
    print_stats_to_run(pairs_to_run)
    for repo, stat in pairs_to_run:
        run_single(repo, stat)


def print_stats_to_run(repo_stats_to_run):
    console = Console()
    table = Table("Repo", "Stat", title="Running the following:", show_lines=True)
    for repo_name, stat_name in repo_stats_to_run:
        table.add_row(f"[green]{repo_name}[/green]", f"[yellow]{stat_name}[/yellow]")
    console.print(table)


def run_single(repo_name: str, stat_name: str):
    print(f"Running [yellow]{stat_name}[/yellow] on [green]{repo_name}[/green]")

    repo_config, stat_params = get_stat_config(repo_name, stat_name)

    stat = Stat(repo_config=repo_config, stat_params=stat_params)
    df = stat.run()



from datetime import datetime
import sh
import shutil
import os
import sys
import logging
import re
import subprocess
from .config import RepoConfig
from . import config
from rich.console import Console
from typing import Optional
from pathlib import PurePath
import pandas as pd

# logging.basicConfig(level=logging.INFO)


# This check was added when the stat runner was cding around too much
# I removed it so that we could dogfood repotracer by running it on itself
def check():
    pass
    revparse = git("rev-parse", "--show-toplevel").strip()
    in_repos_dir = PurePath(os.getcwd()).is_relative_to(config.get_repos_dir())
    if revparse.endswith("repotracer") and not in_repos_dir:
        raise Exception(
            "(bug in repotracer): cannot run a git command against the repotracer repo itself"
        )


git = sh.git.bake(no_pager=True)
git_normal = sh.git.bake(_out=sys.stdout, _err=sys.stderr)


def list_commits(start, end):  # -> [string, string]:
    start = start or first_commit_date()
    end = end or datetime.now().strftime("%Y-%m-%d")
    data = []
    for line in git.log(
        format="%H,%cd",
        date="iso-strict",
        since=start,
        until=end,
        no_merges=True,
        _iter=True,
    ):
        data.append(line.split(","))
    return data


def build_commit_df(
    start: Optional[str], end: Optional[str], freq: str = "D"
) -> pd.DataFrame:
    commits = pd.DataFrame(list_commits(start, end), columns=["sha", "created_at"])
    commits.created_at = pd.DatetimeIndex(
        data=pd.to_datetime(commits.created_at, utc=True)
    )
    commits = commits.set_index(
        commits.created_at,
        drop=False,
    )
    commits = commits.groupby(pd.Grouper(key="created_at", freq=freq)).last()
    return commits


def first_commit_date():
    check()
    # from https://stackoverflow.com/a/5189296
    # Note that that there might be more than one "root" commit
    git_results = git("rev-list", "--max-parents=0", "HEAD").strip()
    # For now we assume that the last one in the list is the earliest, it worked
    # for torvalds/linux
    first_sha = git_results.split("\n")[-1]

    return git.log(first_sha, "--pretty=format:%cd", "--date=format:%Y-%m-%d")


def checkout(sha):
    check()
    return git.checkout(sha)


def reset_hard(target="HEAD"):
    check()
    return git.reset("--hard", target)


def clean_untracked():
    check()
    return git.clean("-fxd")


def current_message():
    check()
    return git.log("-1", "--pretty=%B")


def current_files():
    check()
    return git.diff("--name-only", "HEAD~")


def current_date():
    check()
    return git.log("-1", "--pretty=format:%cd")


def pull(obj="master"):
    check()
    return git.pull("origin", obj)


def get_commit_author():
    check()
    return git.log("-1", "--pretty=format:%aE")


def is_repo_setup(repo_path):
    return os.path.exists(os.path.join(repo_path, ".git"))


def get_default_branch():
    # There is no 'canonical' method of getting the default branch
    # cf https://stackoverflow.com/questions/28666357/how-to-get-default-git-branch
    # This is similar to the sed answer ^
    git_output = git("remote", "show", "origin")
    print(git_output)
    try:
        res = re.search("HEAD branch: (.*)", git_output).group(1)
    except Error:
        print(
            "Could not find default branch, guessing 'master'. If this is incorrect, set the default_branch key on the repo in config.json"
        )
        res = "master"
    return res


def is_url(x):
    try:
        result = urlparse(x)
        return all([result.scheme, result.netloc])
    except:
        return False

    print("default branch is ", default_branch, " checking it out")
    """
            # todo make a more full featured version of this in the git module
        source_path = os.path.expanduser(url_or_path)
        repo_name = name or os.path.basename(source_path)
        repo_storage_path = os.path.join("./repos", repo_name)
        repo_git_dir = os.path.join(repo_storage_path, ".git")
        cwd = os.getcwd()
        os.makedirs(repo_git_dir, exist_ok=True)
        git_dir = os.path.join(source_path, ".git")
        with console.status(f"Copying {git_dir} into '{repo_storage_path}'..."):
            shutil.copytree(git_dir, repo_git_dir, dirs_exist_ok=True)
        os.chdir(repo_storage_path)
        git.checkout(".")
        default_branch = git.get_default_branch()
        git.checkout(default_branch)
        os.chdir(cwd)
        repo_config = config.RepoConfig(
            name=repo_name, path=repo_name, default_branch=default_branch
        )


"""


def download_repo(url_or_path, branch=None):
    cwd = os.getcwd()

    if is_url(url_or_path):
        # splitext to remove the .git off the end
        repo_name = os.path.splitext(os.path.basename(url_or_path))[0]
        repo_source_path = url_or_path
    else:
        # possibly add expanduser here?
        repo_source_path = os.path.abspath(url_or_path)
        repo_name = os.path.basename(repo_source_path)

    repo_storage_path = os.path.join(config.get_repos_dir(), repo_name)
    if os.path.exists(repo_storage_path):
        print(f"Repo {repo_name} already downloaded into {repo_storage_path}")
        return RepoConfig(
            name=repo_name,
            source=repo_source_path,
            storage_path=None,
            default_branch=None,
        )
    os.makedirs(repo_storage_path, exist_ok=True)
    print(f"Cloning {repo_name} from {repo_source_path} to {repo_storage_path}")
    os.chdir(config.get_repos_dir())

    if is_url(url_or_path):
        git_normal.clone(repo_source_path, repo_name, "--single-branch")
    else:
        source_git_dir = os.path.join(repo_source_path, ".git")
        target_git_dir = os.path.join(repo_storage_path, ".git")
        os.makedirs(target_git_dir, exist_ok=True)
        # with console.status(f"Copying {source_git_dir} into '{target_git_dir}'..."):
        print(f"Copying {source_git_dir} into '{target_git_dir}'...")
        shutil.copytree(source_git_dir, target_git_dir, dirs_exist_ok=True)
        os.chdir(repo_storage_path)
        git.checkout(".")
        print(git.status())

    default_branch = get_default_branch()
    git.checkout(default_branch)

    os.chdir(cwd)
    return RepoConfig(
        name=repo_name,
        source=repo_source_path,
        storage_path=None,
        default_branch=default_branch,
    )



from dataclasses import dataclass
from typing import Callable, Generic, TypeVar, Type, Any, TypedDict
from abc import ABC, abstractmethod
import pandas as pd
from tqdm.auto import tqdm
from datetime import datetime
import functools
from typing_extensions import Protocol
from .measurement_fns import rg_count, fd_count, tokei_count, script_auto
from collections import namedtuple


class MeasurementConfig(object):
    pass


TConfig = TypeVar("TConfig", bound=MeasurementConfig)


class Measurement(Protocol):
    def __call__(self, Tconfig) -> list[Any]:
        pass


class FunctionMeasurement(Generic[TConfig]):
    """A Measurement that consists of a single function.
    Easy to turn a function into a measurement."""

    def __init__(self, fn: Callable[[TConfig], list[Any]]):
        self.fn = fn

    def __call__(self, config: TConfig):
        def wrapper() -> list[Any]:
            return self.fn(config)

        return wrapper


class RegexConfig(TypedDict):
    pattern: str


class FileCountConfig(TypedDict):
    pattern: str


class LOCCountConfig(TypedDict):
    pattern: str


class ScriptConfig(TypedDict):
    pattern: str


RegexMeasurement = FunctionMeasurement[RegexConfig](
    lambda config: rg_count(config["pattern"], config.get("ripgrep_args"))
)

FileCountMeasurement = FunctionMeasurement[FileCountConfig](
    lambda config: fd_count(config["pattern"], config.get("fd_args"))
)

LocCountMeasurement = FunctionMeasurement[LOCCountConfig](
    lambda config: tokei_count(config.get("languages"), config.get("total"))
)

ScriptMeasurement = FunctionMeasurement[ScriptConfig](
    lambda config: script_auto(config["command"], "number")
)
# jsx_to_tsx = FunctionMeasurement(tokei_specific(["jsx", "tsx", "js", "ts"]))
# authors_per_month = FunctionMeasurement(git.get_commit_author)

ParamOption = namedtuple("ParamOption", ["name", "description", "required"])
MeasurementDef = namedtuple("MeasurementDef", ["obj", "params"])

all_measurements = {
    "regex_count": MeasurementDef(
        obj=RegexMeasurement,
        params=[
            ParamOption(
                name="pattern",
                description="The regex pattern to pass to ripgrep",
                required=True,
            ),
            ParamOption(
                name="ripgrep_args",
                description="(Optional) any additional ripgrep args. Useful to exclude files with -g ",
                required=False,
            ),
        ],
    ),
    "file_count": MeasurementDef(
        obj=FileCountMeasurement,
        params=[
            ParamOption(
                name="pattern",
                description="""The glob pattern to pass to fd. eg: '*.py' or 'src/**/*.js' """,
                required=True,
            ),
        ],
    ),
    "loc_count": MeasurementDef(
        obj=LocCountMeasurement,
        params=[
            ParamOption(
                name="languages",
                description="""The languages to pass to tokei. eg: 'jsx,tsx,js,ts' """,
                required=False,
            ),
            ParamOption(
                name="total",
                description="""Whether to sum the total lines of code.""",
                required=False,
            ),
        ],
    ),
    "custom_script": MeasurementDef(
        obj=ScriptMeasurement,
        params=[
            ParamOption(
                name="command",
                description="""A shell command to run, like 'wc -l *', or the path to a script you've written: './my_script.sh'""",
                required=True,
            ),
            ParamOption(
                name="return_type",
                description="""The return type. Either 'number' (the command prints a single number), or 'json' (the command prints a line of JSON).""",
                required=False,
            ),
        ],
    ),
}



import pandas as pd
import os

from .config import get_stats_dir

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())


class Storage(object):
    def __init__(self):
        pass

    def save(self, df: pd.DataFrame):
        pass

    def load(self) -> pd.DataFrame:
        pass


class CsvStorage(Storage):
    def __init__(self):
        pass

    def build_path(self, repo_name, stat_name):
        return os.path.join(get_stats_dir(), f"{repo_name}/{stat_name}.csv")

    def load(self, repo_name, stat_name) -> pd.DataFrame | None:
        path = self.build_path(repo_name, stat_name)
        logger.debug(f"{os.getcwd()}: Loading {stat_name} from {repo_name} from {path}")
        try:
            df = pd.read_csv(path, index_col=0)
            df.index = pd.to_datetime(df.index)
            return df
        except FileNotFoundError:
            logger.debug("df not found")
            return None

    def save(self, repo_name, stat_name, df: pd.DataFrame):
        path = self.build_path(repo_name, stat_name)
        # if the path doesn't exist, create it
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path), exist_ok=True)
        logger.info(f"Saving {stat_name} to {path}")
        df.to_csv(path, date_format="%Y-%m-%d")
